@@ -1,12 +1,8 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright (c) 2014-2021 Pytroll developers
+# Copyright (c) 2014-2024 Pytroll developers
 #
-# Author(s):
-#
-#   Adam.Dybbroe <adam.dybbroe@smhi.se>
-#   Panu Lahtinen <panu.lahtinen@fmi.fi>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -23,57 +19,101 @@
 
 """Utility functions."""
 
-import os
 import logging
-import tempfile
+import os
+import sys
+import tarfile
+import warnings
+from functools import wraps
+from inspect import getfullargspec
+
 import numpy as np
-from pyspectral.config import get_config
+import requests
+from scipy.integrate import trapezoid
+
 from pyspectral.bandnames import BANDNAMES
+from pyspectral.config import get_config
+
+TQDM_LOADED = True
+try:
+    from tqdm import tqdm
+except ImportError:
+    TQDM_LOADED = False
+
 
 LOG = logging.getLogger(__name__)
 
 WAVE_LENGTH = 'wavelength'
 WAVE_NUMBER = 'wavenumber'
 
-
-INSTRUMENTS = {'NOAA-19': 'avhrr/3',
-               'NOAA-18': 'avhrr/3',
-               'NOAA-17': 'avhrr/3',
-               'NOAA-16': 'avhrr/3',
-               'NOAA-15': 'avhrr/3',
-               'NOAA-14': 'avhrr/2',
-               'NOAA-12': 'avhrr/2',
-               'NOAA-11': 'avhrr/2',
-               'NOAA-9': 'avhrr/2',
-               'NOAA-7': 'avhrr/2',
+INSTRUMENTS = {'Envisat': 'aatsr',
+               'GOES-16': 'abi',
+               'GOES-17': 'abi',
+               'GOES-18': 'abi',
+               'GOES-19': 'abi',
+               'FY-4A': 'agri',
+               'FY-4B': ['agri', 'ghi'],
+               'Himawari-8': 'ahi',
+               'Himawari-9': 'ahi',
+               'GEO-KOMPSAT-2A': 'ami',
+               'GEO-KOMPSAT-2B': 'goci-2',
                'NOAA-10': 'avhrr/1',
-               'NOAA-8': 'avhrr/1',
                'NOAA-6': 'avhrr/1',
+               'NOAA-8': 'avhrr/1',
                'TIROS-N': 'avhrr/1',
+               'NOAA-11': 'avhrr/2',
+               'NOAA-12': 'avhrr/2',
+               'NOAA-14': 'avhrr/2',
+               'NOAA-7': 'avhrr/2',
+               'NOAA-9': 'avhrr/2',
                'Metop-A': 'avhrr/3',
                'Metop-B': 'avhrr/3',
                'Metop-C': 'avhrr/3',
-               'Suomi-NPP': 'viirs',
-               'NOAA-20': 'viirs',
+               'NOAA-15': 'avhrr/3',
+               'NOAA-16': 'avhrr/3',
+               'NOAA-17': 'avhrr/3',
+               'NOAA-18': 'avhrr/3',
+               'NOAA-19': 'avhrr/3',
+               'HY-1C': 'cocts',
+               'Meteosat-12': 'fci',
+               'MTG-I1': 'fci',
+               'Metop-SG-A1': 'metimage',
                'EOS-Aqua': 'modis',
                'EOS-Terra': 'modis',
-               'FY-3D': 'mersi-2',
-               'FY-3C': 'virr',
-               'FY-3B': 'virr',
-               'Feng-Yun 3D': 'mersi-2',
-               'Meteosat-11': 'seviri',
+               'Sentinel-2A': 'msi',
+               'Sentinel-2B': 'msi',
+               'Sentinel-2C': 'msi',
+               'Arctica-M-N1': 'msu-gsa',
+               'Electro-L-N2': 'msu-gs',
+               'Sentinel-3A': ['olci', 'slstr'],
+               'Sentinel-3B': ['olci', 'slstr'],
+               'Landsat-8': 'oli_tirs',
+               'Landsat-9': 'oli_tirs',
                'Meteosat-10': 'seviri',
-               'Meteosat-9': 'seviri',
+               'Meteosat-11': 'seviri',
                'Meteosat-8': 'seviri',
-               'FY-4A': 'agri',
-               'GEO-KOMPSAT-2A': 'ami',
-               'MTG-I1': 'fci'
-               }
+               'Meteosat-9': 'seviri',
+               'NOAA-20': 'viirs',
+               'NOAA-21': 'viirs',
+               'Suomi-NPP': 'viirs',
+               'FY-3A': ['virr', 'mersi-1'],
+               'FY-3B': ['virr', 'mersi-1'],
+               'FY-3C': ['virr', 'mersi-1'],
+               'FY-3D': 'mersi-2',
+               'FY-3F': 'mersi-3',
+               'FY-3G': 'mersi-rm',
+               'DSCOVR': 'epic'}
 
-HTTP_PYSPECTRAL_RSR = "https://zenodo.org/record/5797957/files/pyspectral_rsr_data.tgz"
+
+INSTRUMENT_TRANSLATION_DASH2SLASH = {'avhrr-1': 'avhrr/1',
+                                     'avhrr-2': 'avhrr/2',
+                                     'avhrr-3': 'avhrr/3'}
+
+HTTP_PYSPECTRAL_RSR = "https://zenodo.org/records/14008148/files/pyspectral_rsr_data.tgz"
 
 RSR_DATA_VERSION_FILENAME = "PYSPECTRAL_RSR_VERSION"
-RSR_DATA_VERSION = "v1.0.17"
+RSR_DATA_VERSION = "v1.4.1"
+
 
 ATM_CORRECTION_LUT_VERSION = {}
 ATM_CORRECTION_LUT_VERSION['antarctic_aerosol'] = {'version': 'v1.0.1',
@@ -99,6 +139,7 @@ ATM_CORRECTION_LUT_VERSION['urban_aerosol'] = {'version': 'v1.0.1',
 ATM_CORRECTION_LUT_VERSION['rayleigh_only'] = {'version': 'v1.0.1',
                                                'filename': 'PYSPECTRAL_ATM_CORR_LUT_RO'}
 
+#: Aerosol types available as downloadable LUTs for rayleigh correction
 AEROSOL_TYPES = ['antarctic_aerosol', 'continental_average_aerosol',
                  'continental_clean_aerosol', 'continental_polluted_aerosol',
                  'desert_aerosol', 'marine_clean_aerosol',
@@ -110,28 +151,18 @@ ATMOSPHERES = {'subarctic summer': 4, 'subarctic winter': 5,
                'tropical': 8, 'us-standard': 9}
 
 HTTPS_RAYLEIGH_LUTS = {}
-URL_PREFIX = "https://zenodo.org/record/1288441/files/pyspectral_atm_correction_luts"
+LUT_URL_PREFIX = "https://zenodo.org/record/1288441/files/pyspectral_atm_correction_luts"
 for atype in AEROSOL_TYPES:
     name = {'rayleigh_only': 'no_aerosol'}.get(atype, atype)
-    url = "{prefix}_{name}.tgz".format(prefix=URL_PREFIX, name=name)
+    url = "{prefix}_{name}.tgz".format(prefix=LUT_URL_PREFIX, name=name)
     HTTPS_RAYLEIGH_LUTS[atype] = url
 
-CONF = get_config()
-LOCAL_RSR_DIR = CONF.get('rsr_dir')
-LOCAL_RAYLEIGH_DIR = CONF.get('rayleigh_dir')
 
-try:
-    os.makedirs(LOCAL_RSR_DIR)
-except OSError:
-    if not os.path.isdir(LOCAL_RSR_DIR):
-        raise
-
-RAYLEIGH_LUT_DIRS = {}
-for sub_dir_name in HTTPS_RAYLEIGH_LUTS:
-    dirname = os.path.join(LOCAL_RAYLEIGH_DIR, sub_dir_name)
-    RAYLEIGH_LUT_DIRS[sub_dir_name] = dirname
-
-TB2RAD_DIR = CONF.get('tb2rad_dir', tempfile.gettempdir())
+def get_rayleigh_lut_dir(aerosol_type):
+    """Get the rayleigh LUT directory for the specified aerosol type."""
+    conf = get_config()
+    local_rayleigh_dir = conf.get('rayleigh_dir')
+    return os.path.join(local_rayleigh_dir, aerosol_type)
 
 
 def convert2wavenumber(rsr):
@@ -198,13 +229,11 @@ def get_central_wave(wav, resp, weight=1.0):
     # if info['unit'].find('-1') > 0:
     # Wavenumber:
     #     res *=
-    return np.trapz(resp * wav * weight, wav) / np.trapz(resp * weight, wav)
+    return trapezoid(resp * wav * weight, wav) / trapezoid(resp * weight, wav)
 
 
 def get_bandname_from_wavelength(sensor, wavelength, rsr, epsilon=0.1, multiple_bands=False):
     """Get the bandname from h5 rsr provided the approximate wavelength."""
-    # channel_list = [channel for channel in rsr.rsr if abs(
-    # rsr.rsr[channel]['det-1']['central_wavelength'] - wavelength) < epsilon]
     chdist_min = 2.0
     chfound = []
     for channel in rsr:
@@ -244,7 +273,7 @@ def sort_data(x_vals, y_vals):
     return x_vals, y_vals
 
 
-def convert2hdf5(ClassIn, platform_name, bandnames, scale=1e-06):
+def convert2hdf5(ClassIn, platform_name, bandnames, scale=1e-06, detectors=None):
     """Retrieve original RSR data and convert to internal hdf5 format.
 
     *scale* is the number which has to be multiplied to the wavelength data in
@@ -268,120 +297,139 @@ def convert2hdf5(ClassIn, platform_name, bandnames, scale=1e-06):
         for chname in bandnames:
             sensor = ClassIn(chname, platform_name)
             grp = h5f.create_group(chname)
-            wvl = sensor.rsr['wavelength'][~np.isnan(sensor.rsr['wavelength'])]
-            rsp = sensor.rsr['response'][~np.isnan(sensor.rsr['wavelength'])]
-            grp.attrs['central_wavelength'] = get_central_wave(wvl, rsp)
-            arr = sensor.rsr['wavelength']
+
+            # If multiple detectors, assume all have same wavelength range in SRF.
+            if detectors is not None:
+                wvl = sensor.rsr[detectors[0]]['wavelength'][~np.isnan(sensor.rsr[detectors[0]]['wavelength'])]
+                arr = sensor.rsr[detectors[0]]['wavelength']
+                grp.attrs['number_of_detectors'] = len(detectors)
+            else:
+                wvl = sensor.rsr['wavelength'][~np.isnan(sensor.rsr['wavelength'])]
+                arr = sensor.rsr['wavelength']
+
+            # Save wavelengths to file
             dset = grp.create_dataset('wavelength', arr.shape, dtype='f')
             dset.attrs['unit'] = 'm'
             dset.attrs['scale'] = scale
             dset[...] = arr
-            arr = sensor.rsr['response']
-            dset = grp.create_dataset('response', arr.shape, dtype='f')
-            dset[...] = arr
+
+            # Now to do the responses
+            if detectors is None:
+                rsp = sensor.rsr['response'][~np.isnan(sensor.rsr['wavelength'])]
+                grp.attrs['central_wavelength'] = get_central_wave(wvl, rsp)
+                arr = sensor.rsr['response']
+                dset = grp.create_dataset('response', arr.shape, dtype='f')
+                dset[...] = arr
+            else:
+                for cur_det in detectors:
+                    det_grp = grp.create_group(cur_det)
+                    rsp = sensor.rsr[cur_det]['response'][~np.isnan(sensor.rsr[cur_det]['wavelength'])]
+                    det_grp.attrs['central_wavelength'] = get_central_wave(wvl, rsp)
+                    arr = sensor.rsr[cur_det]['response']
+                    dset = det_grp.create_dataset('response', arr.shape, dtype='f')
+                    dset[...] = arr
 
 
-def download_rsr(**kwargs):
+def download_rsr(dest_dir=None, dry_run=False):
     """Download the relative spectral response functions.
 
-    Download the pre-compiled hdf5 formatet relative spectral response functions
-    from the internet
+    Download the pre-compiled HDF5 formatted relative spectral response
+    functions from the internet as tarballs, extracts them, then deletes
+    the tarball.
+
+    See :func:`pyspectral.rsr_reader.check_and_download` for a "smart" version
+    of this process that only downloads the necessary files.
+
+    Args:
+        dest_dir (str): Path to put the temporary tarball and extracted RSR
+            files.
+        dry_run (bool): If True, don't actually download files, only log what
+            URLs would be downloaded. Defaults to False.
 
     """
-    import tarfile
-    import requests
-    TQDM_LOADED = True
-    try:
-        from tqdm import tqdm
-    except ImportError:
-        TQDM_LOADED = False
-
-    dest_dir = kwargs.get('dest_dir', LOCAL_RSR_DIR)
-    dry_run = kwargs.get('dry_run', False)
+    config = get_config()
+    local_rsr_dir = config.get('rsr_dir')
+    dest_dir = dest_dir or local_rsr_dir
 
     LOG.info("Download RSR files and store in directory %s", dest_dir)
-
     filename = os.path.join(dest_dir, "pyspectral_rsr_data.tgz")
-    LOG.debug("Get data. URL: %s", HTTP_PYSPECTRAL_RSR)
+    LOG.debug("RSR URL: %s", HTTP_PYSPECTRAL_RSR)
     LOG.debug("Destination = %s", dest_dir)
     if dry_run:
         return
 
-    response = requests.get(HTTP_PYSPECTRAL_RSR)
-    if TQDM_LOADED:
-        with open(filename, "wb") as handle:
-            for data in tqdm(response.iter_content()):
-                handle.write(data)
-    else:
-        with open(filename, "wb") as handle:
-            for data in response.iter_content():
-                handle.write(data)
-
-    tar = tarfile.open(filename)
-    tar.extractall(dest_dir)
-    tar.close()
-    os.remove(filename)
+    _download_tarball_and_extract(HTTP_PYSPECTRAL_RSR, filename, dest_dir)
 
 
-def download_luts(**kwargs):
-    """Download the luts from internet."""
-    import tarfile
-    import requests
-    TQDM_LOADED = True
-    try:
-        from tqdm import tqdm
-    except ImportError:
-        TQDM_LOADED = False
+def download_luts(aerosol_types=None, dry_run=False, aerosol_type=None):
+    """Download the luts from internet.
 
-    dry_run = kwargs.get('dry_run', False)
+    See :func:`pyspectral.rayleigh.check_and_download` for a "smart" version
+    of this process that only downloads the necessary files.
 
-    if 'aerosol_type' in kwargs:
-        if isinstance(kwargs['aerosol_type'], (list, tuple, set)):
-            aerosol_types = kwargs['aerosol_type']
-        else:
-            aerosol_types = [kwargs['aerosol_type'], ]
-    else:
-        aerosol_types = HTTPS_RAYLEIGH_LUTS.keys()
-
-    chunk_size = 1024 * 1024  # 1 MB
+    Args:
+        aerosol_types (Iterable): Aerosol types to download the LUTs for.
+            Defaults to all aerosol types. See :data:`AEROSOL_TYPES` for the
+            full list.
+        dry_run (bool): If True, don't actually download files, only log what
+            URLs would be downloaded. Defaults to False.
+        aerosol_type (str): Deprecated.
 
+    """
+    aerosol_types = _get_aerosol_types(aerosol_types, aerosol_type)
     for subname in aerosol_types:
-
         LOG.debug('Aerosol type: %s', subname)
-        http = HTTPS_RAYLEIGH_LUTS[subname]
-        LOG.debug('URL = %s', http)
-
-        subdir_path = RAYLEIGH_LUT_DIRS[subname]
-        try:
-            LOG.debug('Create directory: %s', subdir_path)
-            if not dry_run:
-                os.makedirs(subdir_path)
-        except OSError:
-            if not os.path.isdir(subdir_path):
-                raise
+        lut_tarball_url = HTTPS_RAYLEIGH_LUTS[subname]
+        LOG.debug('Atmospheric LUT URL = %s', lut_tarball_url)
 
+        subdir_path = get_rayleigh_lut_dir(subname)
+        LOG.debug('Create directory: %s', subdir_path)
+        if not dry_run:
+            os.makedirs(subdir_path, exist_ok=True)
         if dry_run:
             continue
 
-        response = requests.get(http)
-        total_size = int(response.headers['content-length'])
+        local_tarball_pathname = os.path.join(subdir_path, "pyspectral_rayleigh_correction_luts.tgz")
+        _download_tarball_and_extract(lut_tarball_url, local_tarball_pathname, subdir_path)
 
-        filename = os.path.join(
-            subdir_path, "pyspectral_rayleigh_correction_luts.tgz")
-        if TQDM_LOADED:
-            with open(filename, "wb") as handle:
-                for data in tqdm(iterable=response.iter_content(chunk_size=chunk_size),
-                                 total=(int(total_size / chunk_size + 0.5)), unit='kB'):
-                    handle.write(data)
+
+def _get_aerosol_types(aerosol_types, aerosol_type):
+    if aerosol_type is not None:
+        warnings.warn("'aerosol_type' is deprecated, use 'aerosol_types' instead.", UserWarning,
+                      stacklevel=3)
+        if isinstance(aerosol_type, (list, tuple, set)):
+            aerosol_types = aerosol_type
         else:
-            with open(filename, "wb") as handle:
-                for data in response.iter_content():
-                    handle.write(data)
+            aerosol_types = [aerosol_type]
+    elif aerosol_types is None:
+        aerosol_types = list(HTTPS_RAYLEIGH_LUTS.keys())
+    return aerosol_types
 
-        tar = tarfile.open(filename)
-        tar.extractall(subdir_path)
-        tar.close()
-        os.remove(filename)
+
+def _download_tarball_and_extract(tarball_url, local_pathname, extract_dir):
+    chunk_size = 1024 * 1024  # 1 MB
+    response = requests.get(tarball_url)
+    total_size = int(response.headers['content-length'])
+
+    with open(local_pathname, "wb") as handle:
+        for data in _tqdm_or_iter(response.iter_content(chunk_size=chunk_size),
+                                  total=(int(total_size / chunk_size + 0.5)),
+                                  unit='kB'):
+            handle.write(data)
+
+    tar = tarfile.open(local_pathname)
+    tar_kwargs = {} if sys.version_info < (3, 12) else {"filter": "data"}
+    tar.extractall(extract_dir, **tar_kwargs)
+    tar.close()
+    os.remove(local_pathname)
+
+
+def _tqdm_or_iter(an_iterable, **tqdm_kwargs):
+    """Wrap an iterable with tqdm if it is available, otherwise return the iterable."""
+    if TQDM_LOADED:
+        return tqdm(iterable=an_iterable, **tqdm_kwargs)
+    else:
+        return an_iterable
 
 
 def debug_on():
@@ -465,20 +513,23 @@ def convert2str(value):
 
 
 def np2str(value):
-    """Convert an `numpy.string_` to str.
+    """Convert an ``numpy.bytes_`` to str.
+
+    Note: ``numpy.string_`` was deprecated in numpy 2.0 in favor of
+    ``numpy.bytes_``.
 
     Args:
         value (ndarray): scalar or 1-element numpy array to convert
     Raises:
         ValueError: if value is array larger than 1-element or it is not of
-                    type `numpy.string_` or it is not a numpy array
+                    type `numpy.bytes_` or it is not a numpy array
 
     """
     if isinstance(value, str):
         return value
 
     if hasattr(value, 'dtype') and \
-            issubclass(value.dtype.type, (np.str_, np.string_, np.object_)) \
+            issubclass(value.dtype.type, (np.str_, np.bytes_, np.object_)) \
             and value.size == 1:
         value = value.item()
         # python 3 - was scalar numpy array of bytes
@@ -495,3 +546,84 @@ def bytes2string(var):
     if isinstance(var, bytes):
         return var.decode('utf-8')
     return var
+
+
+def _replace_inst_name(instruments):
+    """Replace an instrument name if it's not available on a platform."""
+    if isinstance(instruments, list):
+        return instruments[0]
+    else:
+        return instruments
+
+
+def check_and_adjust_instrument_name(platform_name, instrument):
+    """Check instrument name and try fix if inconsistent.
+
+    It checks against the possible listed instrument names for each platform.
+    It also makes an adjustment replacing names like avhrr/1 with avhrr1,
+    removing the '/'.
+    """
+    instr = INSTRUMENTS.get(platform_name, instrument.lower())
+    if isinstance(instr, list):
+        for inst in instr:
+            goodinst = are_instruments_identical(inst, instrument.lower())
+            if goodinst:
+                break
+    else:
+        goodinst = are_instruments_identical(instr, instrument.lower())
+
+    if not goodinst:
+        instrument = _replace_inst_name(instr)
+        LOG.warning("Inconsistent instrument/satellite input - instrument set to %s",
+                    instrument)
+
+    return instrument.lower().replace('/', '').replace('-', '')
+
+
+def are_instruments_identical(name1, name2):
+    """Given two instrument names check if they are both describing the same instrument.
+
+    Takes care of the case of AVHRR where the internal pyspectral naming
+    (following WMO Oscar) is is with a slash as in 'avhrr/1', but where a
+    naming using a dash instead is equally accepted, as in 'avhrr-1'.
+    """
+    if not isinstance(name1, str) or not isinstance(name2, str):
+        raise ValueError("Instrument names must be strings.")
+
+    if name1 == name2:
+        return True
+    if name1 == INSTRUMENT_TRANSLATION_DASH2SLASH.get(name2):
+        return True
+    return False
+
+
+def use_map_blocks_on(argument_to_run_map_blocks_on):
+    """Use map blocks on a given argument.
+
+    This decorator assumes only one of the arguments of the decorated function is chunked.
+    """
+    def decorator(f):
+        argspec = getfullargspec(f)
+        argument_index = argspec.args.index(argument_to_run_map_blocks_on)
+
+        @wraps(f)
+        def wrapper(*args, **kwargs):
+            array = args[argument_index]
+            chunks = getattr(array, "chunks", None)
+            if chunks is None:
+                return f(*args, **kwargs)
+            import dask.array as da
+            import xarray as xr
+            if isinstance(array, da.Array):
+                return da.map_blocks(f, *args, **kwargs)
+            elif isinstance(array, xr.DataArray):
+                new_array = array.copy()
+                new_args = list(args)
+                new_args[argument_index] = array.data
+                new_data = da.map_blocks(f, *new_args, **kwargs)
+                new_array.data = new_data
+                return new_array
+            else:
+                raise NotImplementedError(f"Don't know how to map_blocks on {type(array)}")
+        return wrapper
+    return decorator
